#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>

//  レジスタブロックの物理アドレス
//#define BCM2708_PERI_BASE        0x20000000
#define BCM2708_PERI_BASE        0x3F000000
#define GPIO_BASE                (BCM2708_PERI_BASE + 0x200000) /* GPIO   */
#define PADS_BASE                (BCM2708_PERI_BASE + 0x100000) /* PADS   */
#define BLOCK_SIZE    4096

//  ピン機能（BCM2835）
#define GPIO_INPUT    0x0       //  入力
#define GPIO_OUTPUT   0x1       //  出力
#define GPIO_ALT0     0x4
#define GPIO_ALT1     0x5
#define GPIO_ALT2     0x6
#define GPIO_ALT3     0x7
#define GPIO_ALT4     0x3
#define GPIO_ALT5     0x2

#define PADS0   *(pads+11)
#define PADS1   *(pads+12)
#define PADS2   *(pads+13)

// I/O access - Reqires maping.
volatile unsigned *pads = NULL;

void gpio_init(void);

int main(int argc, char *argv[])
{
    gpio_init();

// 0 :  2mA
// 1 :  4mA
// 2 :  6mA
// 3 :  8mA
// 4 : 10mA
// 5 : 12mA
// 6 : 14mA
// 7 : 16mA
// PADS0 : GPIO 0 -27
// PADS1 : GPIO 28-45
// PADS2 : GPIO 46-53
// 0x5a000000 : write password
// BCLK : GPIO 18
// LRCK : GPIO 19
// DOUT : GPIO 21
// I2C  : CPIO 2, 3
//    PADS1 = (0xFFFFFFF8 & PADS1) | 0x7 | 0x5a000000;
    printf("PADS0: %08X (%08X)\n",PADS0, PADS0&0x7);
    printf("PADS1: %08X (%08X)\n",PADS1, PADS1&0x7);
    printf("PADS2: %08X (%08X)\n",PADS2, PADS2&0x7);

    int vol;
    if (argc > 1) {
        vol = atoi(argv[1]);
        PADS0 = (0xFFFFFFF8 & PADS0) | vol | 0x5a000000;
        printf("PADS0: %08X (%08X)\n",PADS0, PADS0&0x7);
    }


    return 0;
}

void gpio_init(void)
{
    //  既に初期化済なら何もしない
    if (pads) return;

    //  ここから GPIO 初期化
    int fd;
    void *pad_map;

    //  /dev/mem（物理メモリデバイス）を開く（sudo が必要）
    fd = open("/dev/mem", O_RDWR | O_SYNC);

    if (fd == -1) {
        printf("error: cannot open /dev/mem (gpio_setup)n");
        exit(-1);
    }

    //  mmapで物理メモリを仮想メモリに対応づける
    pad_map = mmap(NULL, BLOCK_SIZE,
                    PROT_READ | PROT_WRITE, MAP_SHARED,
                    fd, PADS_BASE );

    if ((int) pad_map == -1) {
        printf("error: cannot map /dev/mem on the memory (gpio_setup)n");
        exit(-1);
    }

    //  mmap 後は不要な fd をクローズ
    close(fd);

    pads = (unsigned int *) pad_map;
}


